#include <stdlib.h>
#include <string.h>

#include <compiz.h>

#define OPACITY_STEP_DEFAULT	10
#define OPACITY_STEP_MIN	1
#define OPACITY_STEP_MAX	50

#define OPACITY_INCREASE_BUTTON_DEFAULT 	Button4
#define OPACITY_INCREASE_MODIFIERS_DEFAULT	(CompPressMask | ControlMask | ShiftMask)

#define OPACITY_DECREASE_BUTTON_DEFAULT 	Button5
#define OPACITY_DECREASE_MODIFIERS_DEFAULT	(CompPressMask | ControlMask | ShiftMask)

static char *winType[] = {
    "Dock",
    "Toolbar",
    "Menu",
    "Utility",
    "Splash",
    "Normal",
    "Dialog",
    "ModalDialog",
    "Unknown"
};
#define N_WIN_TYPE (sizeof (winType) / sizeof (winType[0]))

static int displayPrivateIndex;

typedef struct _OpacityDisplay {
    int		    screenPrivateIndex;
    HandleEventProc handleEvent;
} OpacityDisplay;

#define OPACITY_SCREEN_OPTION_STEP		0
#define OPACITY_SCREEN_OPTION_INCREASE		1
#define OPACITY_SCREEN_OPTION_DECREASE		2
#define OPACITY_SCREEN_OPTION_WINDOW_TYPE	3
#define OPACITY_SCREEN_OPTION_NUM		4

typedef struct _OpacityScreen {
    int			   opacityStep;

    CompOption opt[OPACITY_SCREEN_OPTION_NUM];

    int wMask;
} OpacityScreen;

#define GET_OPACITY_DISPLAY(d)				     \
    ((OpacityDisplay *) (d)->privates[displayPrivateIndex].ptr)

#define OPACITY_DISPLAY(d)			   \
    OpacityDisplay *fd = GET_OPACITY_DISPLAY (d)

#define GET_OPACITY_SCREEN(s, fd)					 \
    ((OpacityScreen *) (s)->privates[(fd)->screenPrivateIndex].ptr)

#define OPACITY_SCREEN(s)							\
    OpacityScreen *fs = GET_OPACITY_SCREEN (s, GET_OPACITY_DISPLAY (s->display))

#define NUM_OPTIONS(s) (sizeof ((s)->opt) / sizeof (CompOption))

static CompOption *
opacityGetScreenOptions (CompScreen *screen,
		      int	 *count)
{
    OPACITY_SCREEN (screen);

    *count = NUM_OPTIONS (fs);
    return fs->opt;
}

static Bool
opacitySetScreenOption (CompScreen      *screen,
		     char	     *name,
		     CompOptionValue *value)
{
    CompOption *o;
    int	       index;

    OPACITY_SCREEN (screen);

    o = compFindOption (fs->opt, NUM_OPTIONS (fs), name, &index);
    if (!o)
	return FALSE;

    switch (index) {
    case OPACITY_SCREEN_OPTION_STEP:
	if (compSetIntOption (o, value))
	{
	    fs->opacityStep = o->value.i;
	    return TRUE;
	}
	break;
    case OPACITY_SCREEN_OPTION_INCREASE:
    case OPACITY_SCREEN_OPTION_DECREASE:
	if (addScreenBinding (screen, &value->bind))
	{
	    removeScreenBinding (screen, &o->value.bind);

	    if (compSetBindingOption (o, value))
		return TRUE;
	}
	break;
    case OPACITY_SCREEN_OPTION_WINDOW_TYPE:
	if (compSetOptionList (o, value))
	{
	    fs->wMask = compWindowTypeMaskFromStringList (&o->value);
	    fs->wMask &= ~CompWindowTypeDesktopMask;
	    return TRUE;
	}
    default:
	break;
    }

    return FALSE;
}

static void
opacityScreenInitOptions (OpacityScreen *fs)
{
    CompOption *o;
    int	       i;

    o = &fs->opt[OPACITY_SCREEN_OPTION_STEP];
    o->name		= "step";
    o->shortDesc	= "Opacity Step";
    o->longDesc		= "Opacity change step";
    o->type		= CompOptionTypeInt;
    o->value.i		= OPACITY_STEP_DEFAULT;
    o->rest.i.min	= OPACITY_STEP_MIN;
    o->rest.i.max	= OPACITY_STEP_MAX;

    o = &fs->opt[OPACITY_SCREEN_OPTION_INCREASE];
    o->name			     = "increase";
    o->shortDesc		     = "Increase Opacity";
    o->longDesc			     = "Increase Opacity";
    o->type			     = CompOptionTypeBinding;
    o->value.bind.type		     = CompBindingTypeButton;
    o->value.bind.u.button.modifiers = OPACITY_INCREASE_MODIFIERS_DEFAULT;
    o->value.bind.u.button.button    = OPACITY_INCREASE_BUTTON_DEFAULT;

    o = &fs->opt[OPACITY_SCREEN_OPTION_DECREASE];
    o->name			     = "decrease";
    o->shortDesc		     = "Decrease Opacity";
    o->longDesc			     = "Decrease Opacity";
    o->type			     = CompOptionTypeBinding;
    o->value.bind.type		     = CompBindingTypeButton;
    o->value.bind.u.button.modifiers = OPACITY_DECREASE_MODIFIERS_DEFAULT;
    o->value.bind.u.button.button    = OPACITY_DECREASE_BUTTON_DEFAULT;

    o = &fs->opt[OPACITY_SCREEN_OPTION_WINDOW_TYPE];
    o->name	         = "window_types";
    o->shortDesc         = "Window Types";
    o->longDesc	         = "Window types that can have opacity changed";
    o->type	         = CompOptionTypeList;
    o->value.list.type   = CompOptionTypeString;
    o->value.list.nValue = N_WIN_TYPE;
    o->value.list.value  = malloc (sizeof (CompOptionValue) * N_WIN_TYPE);
    for (i = 0; i < N_WIN_TYPE; i++)
	o->value.list.value[i].s = strdup (winType[i]);
    o->rest.s.string     = windowTypeString;
    o->rest.s.nString    = nWindowTypeString;

    fs->wMask = compWindowTypeMaskFromStringList (&o->value);
}

static void
opacityChangeOpacity (CompScreen* s,
		Bool increase)
{
    CompWindow* w;

    OPACITY_SCREEN (s);
    
    if (s->display->activeWindow)
    {
        w = findWindowAtScreen (s, s->display->activeWindow);
    
        if (w && (fs->wMask & w->type))
	{
	    GLushort step = (GLushort)(OPAQUE*(fs->opacityStep / 100.0f));

            if (increase)
	    {
    	        if ((w->paint.opacity + step) < OPAQUE)
		    w->paint.opacity += step;
		else
		    w->paint.opacity = OPAQUE;
	    }
	    else
	    {
		/* don't let them set opacity to 0 - make step the minimum opacity */
		if (w->paint.opacity >= 2*step)
		    w->paint.opacity -= step;
		else
		    w->paint.opacity = step;
	    }

	    addWindowDamage (w);
	}
    }
}

static void
opacityHandleEvent (CompDisplay *d,
		 XEvent      *event)
{
    CompWindow *w;
    CompScreen *s;

    OPACITY_DISPLAY (d);

    switch (event->type) {
    case KeyPress:
    case KeyRelease:
	s = findScreenAtDisplay (d, event->xkey.root);
	if (s)
	{
	    OPACITY_SCREEN (s);

	    if (EV_KEY (&fs->opt[OPACITY_SCREEN_OPTION_INCREASE], event))
		opacityChangeOpacity(s, TRUE);

	    if (EV_KEY (&fs->opt[OPACITY_SCREEN_OPTION_DECREASE], event))
		opacityChangeOpacity(s, FALSE);
	}
	break;
    case ButtonPress:
    case ButtonRelease:
	s = findScreenAtDisplay (d, event->xbutton.root);
	if (s)
	{
	    OPACITY_SCREEN (s);

	    if (EV_BUTTON (&fs->opt[OPACITY_SCREEN_OPTION_INCREASE], event))
		opacityChangeOpacity(s, TRUE);

	    if (EV_BUTTON (&fs->opt[OPACITY_SCREEN_OPTION_DECREASE], event))
		opacityChangeOpacity(s, FALSE);
	}
	break;
    default:
	break;
    }

    UNWRAP (fd, d, handleEvent);
    (*d->handleEvent) (d, event);
    WRAP (fd, d, handleEvent, opacityHandleEvent);
}

static Bool
opacityInitDisplay (CompPlugin  *p,
		 CompDisplay *d)
{
    OpacityDisplay *fd;

    fd = malloc (sizeof (OpacityDisplay));
    if (!fd)
	return FALSE;

    fd->screenPrivateIndex = allocateScreenPrivateIndex (d);
    if (fd->screenPrivateIndex < 0)
    {
	free (fd);
	return FALSE;
    }

    WRAP (fd, d, handleEvent, opacityHandleEvent);

    d->privates[displayPrivateIndex].ptr = fd;

    return TRUE;
}

static void
opacityFiniDisplay (CompPlugin *p,
		 CompDisplay *d)
{
    OPACITY_DISPLAY (d);

    freeScreenPrivateIndex (d, fd->screenPrivateIndex);

    UNWRAP (fd, d, handleEvent);

    free (fd);
}

static Bool
opacityInitScreen (CompPlugin *p,
		CompScreen *s)
{
    OpacityScreen *fs;

    OPACITY_DISPLAY (s->display);

    fs = malloc (sizeof (OpacityScreen));
    if (!fs)
	return FALSE;

    fs->opacityStep = OPACITY_STEP_DEFAULT;

    opacityScreenInitOptions (fs);

    s->privates[fd->screenPrivateIndex].ptr = fs;

    return TRUE;
}

static void
opacityFiniScreen (CompPlugin *p,
		CompScreen *s)
{
    OPACITY_SCREEN (s);

    free (fs);
}

static Bool
opacityInit (CompPlugin *p)
{
    displayPrivateIndex = allocateDisplayPrivateIndex ();
    if (displayPrivateIndex < 0)
	return FALSE;

    return TRUE;
}

static void
opacityFini (CompPlugin *p)
{
    if (displayPrivateIndex >= 0)
	freeDisplayPrivateIndex (displayPrivateIndex);
}

CompPluginDep opacityDeps[] = {
    { CompPluginRuleAfter, "decoration" },
    { CompPluginRuleAfter, "fade" },
    { CompPluginRuleAfter, "switcher" }
};

static CompPluginVTable opacityVTable = {
    "opacity",
    "Change Window Opacity",
    "Change window opacity in steps",
    opacityInit,
    opacityFini,
    opacityInitDisplay,
    opacityFiniDisplay,
    opacityInitScreen,
    opacityFiniScreen,
    0, /* InitWindow */
    0, /* FiniWindow */
    0, /* GetDisplayOptions */
    0, /* SetDisplayOption */
    opacityGetScreenOptions,
    opacitySetScreenOption,
    opacityDeps,
    sizeof (opacityDeps) / sizeof (opacityDeps[0])
};

CompPluginVTable *
getCompPluginInfo (void)
{
    return &opacityVTable;
}
